import matplotlib.pyplot as plt
import numpy as np
import rpy2.robjects.conversion as cv
from rpy2.robjects import (
    default_converter,
    FloatVector,
    ListVector,
    numpy2ri,
    r,
)

from ..Base import Base
from ..utils import multivariate as mv
from ..utils import unimultivariate as umv
from .. import config


class Ridge2Regressor(Base):
    """Random Vector functional link network model with 2 regularization parameters

    Parameters:

        h: an integer;
            forecasting horizon

        level: an integer;
            Confidence level for prediction intervals

        lags: an integer;
            Number of lags

        nb_hidden: an integer;
            Number of nodes in hidden layer

        nodes_sim: an integer;
            Type of simulation for nodes in the hidden layer
            ("sobol", "halton", "unif")

        activation: a string;
            Activation function ("relu", "sigmoid", "tanh",
            "leakyrelu", "elu", "linear")

        a: a float;
            hyperparameter for activation function "leakyrelu", "elu"

        lambda_1: a float;
            Regularization parameter for original predictors

        lambda_2: a float;
            Regularization parameter for transformed predictors

        dropout: a float;
            dropout regularization parameter (dropping nodes in hidden layer)

        type_pi: a string;
            Type of prediction interval (currently "gaussian",
            "bootstrap", (circular) "blockbootstrap", "movingblockbootstrap", or "rvinecopula")

        block_length: an integer
            length of block for multivariate block bootstrap (`type_pi == blockbootstrap` or
            `type_pi == movingblockbootstrap`)

        margins: a string;
            distribution of residuals' marginals for `type_pi == rvinecopula`: "empirical" (default),
            "gaussian"

        B: an integer;
            Number of bootstrap replications for `type_pi == bootstrap`, "blockbootstrap",
            "movingblockbootstrap", or "rvinecopula"

        type_aggregation: a string;
            Type of aggregation, ONLY for bootstrapping; either "mean" or "median"

        centers: an integer;
            Number of clusters for \code{type_clustering}

        type_clustering: a string;
            "kmeans" (K-Means clustering) or "hclust" (Hierarchical clustering)

        cl: an integer;
            The number of clusters for parallel execution (done in R), for `type_pi == bootstrap`

        date_formatting: a string;
            Currently:
            - "original": yyyy-mm-dd
            - "ms": milliseconds

        seed: an integer;
            reproducibility seed for type_pi == 'bootstrap'

    Attributes:

        fcast_: an object;
            raw result from fitting R's `ahead::ridge2f` through `rpy2`

        averages_: a list of lists;
            mean forecast in a list for each series

        ranges_: a list of lists;
            lower and upper prediction intervals in a list for each series

        output_dates_: a list;
            a list of output dates (associated to forecast)

        mean_: a numpy array
            contains series mean forecast as a numpy array

        lower_: a numpy array
            contains series lower bound forecast as a numpy array

        upper_: a numpy array
            contains series upper bound forecast as a numpy array

        result_dfs_: a tuple of data frames;
            each element of the tuple contains 3 columns,
            mean forecast, lower + upper prediction intervals,
            and a date index

        sims_: currently a tuple of numpy arrays
            for `type_pi == bootstrap`, simulations for each series

    Examples:

    ```python
    import pandas as pd
    from ahead import Ridge2Regressor

    # Data frame containing the time series
    dataset = {
    'date' : ['2001-01-01', '2002-01-01', '2003-01-01', '2004-01-01', '2005-01-01'],
    'series1' : [34, 30, 35.6, 33.3, 38.1],
    'series2' : [4, 5.5, 5.6, 6.3, 5.1],
    'series3' : [100, 100.5, 100.6, 100.2, 100.1]}
    df = pd.DataFrame(dataset).set_index('date')
    print(df)

    # multivariate time series forecasting
    r1 = Ridge2Regressor(h = 5)
    r1.forecast(df)
    print(r1.result_dfs_)
    ```

    """

    def __init__(
        self,
        h=5,
        level=95,
        lags=1,
        nb_hidden=5,
        nodes_sim="sobol",
        activation="relu",
        a=0.01,
        lambda_1=0.1,
        lambda_2=0.1,
        dropout=0,
        type_pi="gaussian",
        # can be NULL, but in R (use 0 in R instead of NULL for v0.7.0)
        block_length=3,
        margins="empirical",
        B=100,
        type_aggregation="mean",
        centers=2,
        type_clustering="kmeans",
        cl=1,
        date_formatting="original",
        seed=123,
    ):

        super().__init__(
            h=h,
            level=level,
            seed=seed,
        )

        self.lags = lags
        self.nb_hidden = nb_hidden
        self.nodes_sim = nodes_sim
        self.activation = activation
        self.a = a
        self.lambda_1 = lambda_1
        self.lambda_2 = lambda_2
        self.dropout = dropout
        self.type_pi = type_pi
        self.block_length = block_length
        self.margins = margins
        self.B = B
        self.type_aggregation = type_aggregation
        # can be NULL, but in R (use 0 in R instead of NULL for v0.7.0)
        self.centers = centers
        self.type_clustering = type_clustering
        self.cl = cl
        self.date_formatting = date_formatting
        self.seed = seed
        self.input_df = None
        self.type_input = "multivariate"

        self.fcast_ = None
        self.averages_ = None
        self.ranges_ = None
        self.output_dates_ = []
        self.mean_ = None
        self.lower_ = None
        self.upper_ = None
        self.result_dfs_ = None
        self.sims_ = None
        self.xreg_ = None

    def forecast(self, df, xreg=None):
        """Forecasting method from `Ridge2Regressor` class

        Parameters:

            df: a data frame;
                a data frame containing the input time series (see example)

            xreg: a numpy array or a data frame;
                external regressors

        """

        # get input dates, output dates, number of series, series names, etc. 
        self.init_forecasting_params(df)

        # obtain time series object -----
        self.format_input()

        self.get_forecast("ridge2")

        # result -----
        (
            self.averages_,
            self.ranges_,
            _,
        ) = mv.format_multivariate_forecast(
            n_series=self.n_series,
            date_formatting=self.date_formatting,
            output_dates=self.output_dates_,
            horizon=self.h,
            fcast=self.fcast_,
        )

        self.mean_ = np.asarray(self.fcast_.rx2["mean"])
        self.lower_ = np.asarray(self.fcast_.rx2["lower"])
        self.upper_ = np.asarray(self.fcast_.rx2["upper"])

        self.result_dfs_ = tuple(
            umv.compute_result_df(self.averages_[i], self.ranges_[i])
            for i in range(self.n_series)
        )

        if self.type_pi in (
            "bootstrap",
            "blockbootstrap",
            "movingblockbootstrap",
            "rvinecopula",
        ):
            self.sims_ = tuple(
                np.asarray(self.fcast_.rx2["sims"][i]) for i in range(self.B)
            )

        return self


import numpy as np
import matplotlib.pyplot as plt

from rpy2.robjects import numpy2ri, r
from rpy2.robjects.vectors import FloatVector

from subprocess import Popen, PIPE
from .. import config
from ..utils.multivariate import compute_y_mts
from ..utils.unimultivariate import compute_input_dates, compute_output_dates


class Base(object):

    def __init__(self, h=5, level=95, date_formatting="ms", seed=123):

        self.h = h
        self.level = level
        self.date_formatting = date_formatting
        self.seed = seed
        self.frequency = None 
        self.series_names = None
        self.n_series = None
        self.type_input = "univariate" #(or "multivariate")
        self.B = None
        self.input_df = None
        self.input_dates = None   
        self.method = None  
        self.weights = None      

        self.input_ts_ = None # input time series 
        self.mean_ = None
        self.lower_ = None
        self.upper_ = None
        self.sims_ = None 
        self.output_dates_ = None
        self.result_dfs_ = None 

        R_IS_INSTALLED = False

        try:
            proc = Popen(["which", "R"], stdout=PIPE, stderr=PIPE)
            R_IS_INSTALLED = proc.wait() == 0
        except Exception as e:
            pass

        if not R_IS_INSTALLED:
            raise ImportError("R is not installed! \n" + config.USAGE_MESSAGE)
    
    def format_input(self):
        self.input_ts_ = compute_y_mts(self.input_df, 
                             self.frequency)

    def init_forecasting_params(self, df):
        self.input_df = df                
        self.series_names = df.columns        
        self.n_series = len(self.series_names)        
        self.input_dates = compute_input_dates(df)
        self.type_input = "multivariate" if len(df.shape) > 0 else "univariate"
        self.output_dates_, self.frequency = compute_output_dates(
            df, self.h
        )        

    def getsims(self, input_tuple, ix):
        n_sims = len(input_tuple)
        res = [input_tuple[i].iloc[:, ix].values for i in range(n_sims)]
        return np.asarray(res).T

    def get_forecast(self, method=None, xreg=None):

        if method != None: 
            self.method = method 

        if self.method == "armagarch":
            self.fcast_ = config.AHEAD_PACKAGE.armagarchf(
            y=y,
            h=self.h,
            level=self.level,
            B=self.B,
            cl=self.cl,
            dist=self.dist,
            seed=self.seed)

        if self.method in ("mean", "median", "rw"):
            self.fcast_ = config.AHEAD_PACKAGE.basicf(
            self.input_ts_,
            h=self.h,
            level=self.level,
            method=self.method,
            type_pi=self.type_pi,
            block_length=self.block_length,
            B=self.B,
            seed=self.seed)

        if self.method == "dynrm":
            self.fcast_ = config.AHEAD_PACKAGE.dynrmf(
            y=self.input_ts_, 
            h=self.h, level=self.level, 
            type_pi=self.type_pi)

        if self.method == "eat":
            self.fcast_ = config.AHEAD_PACKAGE.eatf(
            y=self.input_ts_,
            h=self.h,
            level=self.level,
            type_pi=self.type_pi,
            weights=config.FLOATVECTOR(self.weights))

        if self.method == "ridge2":            
            if xreg is None:

                self.fcast_ = config.AHEAD_PACKAGE.ridge2f(
                    self.input_ts_,
                    h=self.h,
                    level=self.level,
                    lags=self.lags,
                    nb_hidden=self.nb_hidden,
                    nodes_sim=self.nodes_sim,
                    activ=self.activation,
                    a=self.a,
                    lambda_1=self.lambda_1,
                    lambda_2=self.lambda_2,
                    dropout=self.dropout,
                    type_pi=self.type_pi,
                    margins=self.margins,
                    # can be NULL, but in R (use 0 in R instead of NULL for v0.7.0)
                    block_length=self.block_length,
                    B=self.B,
                    type_aggregation=self.type_aggregation,
                    # can be NULL, but in R (use 0 in R instead of NULL for v0.7.0)
                    centers=self.centers,
                    type_clustering=self.type_clustering,
                    cl=self.cl,
                    seed=self.seed,
                )

            else:  # xreg is not None:

                try:
                    self.xreg_ = xreg.values
                except:
                    self.xreg_ = config.DEEP_COPY(xreg)

                is_matrix_xreg = len(self.xreg_.shape) > 1

                numpy2ri.activate()

                xreg_ = (
                    r.matrix(
                        FloatVector(self.xreg_.flatten()),
                        byrow=True,
                        nrow=self.xreg_.shape[0],
                        ncol=self.xreg_.shape[1],
                    )
                    if is_matrix_xreg
                    else r.matrix(
                        FloatVector(self.xreg_.flatten()),
                        byrow=True,
                        nrow=self.xreg_.shape[0],
                        ncol=1,
                    )
                )

                self.fcast_ = config.AHEAD_PACKAGE.ridge2f(
                    self.input_ts_,
                    xreg=xreg_,
                    h=self.h,
                    level=self.level,
                    lags=self.lags,
                    nb_hidden=self.nb_hidden,
                    nodes_sim=self.nodes_sim,
                    activ=self.activation,
                    a=self.a,
                    lambda_1=self.lambda_1,
                    lambda_2=self.lambda_2,
                    dropout=self.dropout,
                    type_pi=self.type_pi,
                    margins=self.margins,
                    # can be NULL, but in R (use 0 in R instead of NULL for v0.7.0)
                    block_length=self.block_length,
                    B=self.B,
                    type_aggregation=self.type_aggregation,
                    # can be NULL, but in R (use 0 in R instead of NULL for v0.7.0)
                    centers=self.centers,
                    type_clustering=self.type_clustering,
                    cl=self.cl,
                    seed=self.seed,
                )
        
        if self.method == "var":
            self.fcast_ = config.AHEAD_PACKAGE.varf(
            self.input_ts_,
            h=self.h,
            level=self.level,
            lags=self.lags,
            type_VAR=self.type_VAR)


    def plot(self, series, type_axis="dates", type_plot="pi"):
        """Plot time series forecast

        Parameters:

        series: {integer} or {string}
            series index or name
        """
        assert all(
            [
                self.mean_ is not None,
                self.lower_ is not None,
                self.upper_ is not None,
                self.output_dates_ is not None,
            ]
        ), "model forecasting must be obtained first (with `forecast` method)"

        if isinstance(series, str):
            assert (
                series in self.series_names
            ), f"series {series} doesn't exist in the input dataset"
            series_idx = self.input_df.columns.get_loc(series)
        else:
            assert isinstance(series, int) and (
                0 <= series < self.n_series
            ), f"check series index (< {self.n_series})"
            series_idx = series        

        y_all = list(self.input_df.iloc[:, series_idx]) + list(
            self.result_dfs_[series_idx]["mean"].values
        )        
        
        y_test = list(self.result_dfs_[series_idx]["mean"].values)
        n_points_all = len(y_all)
        n_points_train = self.input_df.shape[0]

        if type_axis == "numeric":
            x_all = [i for i in range(n_points_all)]
            x_test = [i for i in range(n_points_train, n_points_all)]

        if type_axis == "dates":  # use dates    
            x_train = [date.strftime('%Y-%m-%d') for date in self.input_dates]                       
            x_test = [date.strftime('%Y-%m-%d') for date in self.output_dates_]
            x_all = np.concatenate(
                (x_train, x_test), axis=None
            )

        if type_plot == "pi":
            fig, ax = plt.subplots()
            ax.plot(x_all, y_all, "-")
            ax.plot(x_test, y_test, "-", color="orange")
            ax.fill_between(
                x_test,
                self.result_dfs_[series_idx]["lower"].values,
                self.result_dfs_[series_idx]["upper"].values,
                alpha=0.2,
                color="orange",
            )
            plt.title(
                f"prediction intervals for {series}",
                loc="left",
                fontsize=12,
                fontweight=0,
                color="black",
            )
            plt.show()

        if type_plot == "spaghetti":
            palette = plt.get_cmap("Set1")
            sims_ix = self.getsims(self.sims_, series_idx)
            plt.plot(x_all, y_all, "-")
            for col_ix in range(
                sims_ix.shape[1]
            ):  # avoid this when there are thousands of simulations
                plt.plot(
                    x_test,
                    sims_ix[:, col_ix],
                    "-",
                    color=palette(col_ix),
                    linewidth=1,
                    alpha=0.9,
                )
            plt.plot(x_all, y_all, "-", color="black")
            plt.plot(x_test, y_test, "-", color="blue")
            # Add titles
            plt.title(
                f"{self.B} simulations of {series}",
                loc="left",
                fontsize=12,
                fontweight=0,
                color="black",
            )
            plt.xlabel("Time")
            plt.ylabel("Values")
            # Show the graph
            plt.show()



import numpy as np

from ..Base import Base
from ..utils import multivariate as mv
from ..utils import unimultivariate as umv
from .. import config


class BasicForecaster(Base):
    """Basic forecasting functions for multivariate time series (mean, median, random walk)

    Parameters:

        h: an integer;
            forecasting horizon

        level: an integer;
            Confidence level for prediction intervals

        method: a string;
            Forecasting method, either "mean", "median", or random walk ("rw")

        type_pi: a string;
            Type of prediction interval (currently "gaussian",
            "bootstrap" (independent), "blockbootstrap" (circular),
            "movingblockbootstrap")

        block_length: an integer
            length of block for multivariate block bootstrap (`type_pi == blockbootstrap`
            or `type_pi == movingblockbootstrap`)

        B: an integer;
            Number of replications

        date_formatting: a string;
            Currently:
            - "original": yyyy-mm-dd
            - "ms": milliseconds

        seed: an integer;
            reproducibility seed

    Attributes:

        fcast_: an object;
            raw result from fitting R's `ahead::ridge2f` through `rpy2`

        averages_: a list of lists;
            mean forecast in a list for each series

        ranges_: a list of lists;
            lower and upper prediction intervals in a list for each series

        output_dates_: a list;
            a list of output dates (associated to forecast)

        mean_: a numpy array
            contains series mean forecast as a numpy array

        lower_: a numpy array
            contains series lower bound forecast as a numpy array

        upper_: a numpy array
            contains series upper bound forecast as a numpy array

        result_dfs_: a tuple of data frames;
            each element of the tuple contains 3 columns,
            mean forecast, lower + upper prediction intervals,
            and a date index

        sims_: currently a tuple of numpy arrays
            for `type_pi == bootstrap`, simulations for each series

    Examples:

    ```python
    import pandas as pd
    from ahead import BasicForecaster

    # Data frame containing the time series
    dataset = {
    'date' : ['2001-01-01', '2002-01-01', '2003-01-01', '2004-01-01', '2005-01-01'],
    'series1' : [34, 30, 35.6, 33.3, 38.1],
    'series2' : [4, 5.5, 5.6, 6.3, 5.1],
    'series3' : [100, 100.5, 100.6, 100.2, 100.1]}
    df = pd.DataFrame(dataset).set_index('date')
    print(df)

    # multivariate time series forecasting
    r1 = BasicForecaster(h = 5)
    r1.forecast(df)
    print(r1.result_dfs_)
    ```

    """

    def __init__(
        self,
        h=5,
        level=95,
        method="mean",
        type_pi="gaussian",
        block_length=5,
        B=100,
        date_formatting="original",
        seed=123,
    ):

        super().__init__(
            h=h,
            level=level,
            seed=seed,
        )

        self.method = method
        self.type_pi = type_pi
        self.block_length = block_length
        self.B = B
        self.date_formatting = date_formatting
        self.input_df = None        

        self.fcast_ = None
        self.averages_ = None
        self.ranges_ = None
        self.output_dates_ = []
        self.mean_ = None
        self.lower_ = None
        self.upper_ = None
        self.result_dfs_ = None
        self.sims_ = None

    def forecast(self, df):
        """Forecasting method from `BasicForecaster` class

        Parameters:

            df: a data frame;
                a data frame containing the input time series (see example)

        """

        # get input dates, output dates, number of series, series names, etc. 
        self.init_forecasting_params(df)

        # obtain time series object -----
        self.format_input()

        if self.type_pi in ("blockbootstrap", "movingblockbootstrap"):
            assert (
                self.block_length is not None
            ), "For `type_pi in ('blockbootstrap', 'movingblockbootstrap')`, `block_length` must be not None"        

        self.get_forecast()

        # result -----
        (
            self.averages_,
            self.ranges_,
            _,
        ) = mv.format_multivariate_forecast(
            n_series=self.n_series,
            date_formatting=self.date_formatting,
            output_dates=self.output_dates_,
            horizon=self.h,
            fcast=self.fcast_,
        )

        self.mean_ = np.asarray(self.fcast_.rx2["mean"])
        self.lower_ = np.asarray(self.fcast_.rx2["lower"])
        self.upper_ = np.asarray(self.fcast_.rx2["upper"])

        self.result_dfs_ = tuple(
            umv.compute_result_df(self.averages_[i], self.ranges_[i])
            for i in range(self.n_series)
        )

        if self.type_pi in (
            "bootstrap",
            "blockbootstrap",
            "movingblockbootstrap",
        ):
            self.sims_ = tuple(
                np.asarray(self.fcast_.rx2["sims"][i]) for i in range(self.B)
            )

        return self



"""Top-level package for ahead."""

__author__ = """T. Moudiki"""
__email__ = "thierry.moudiki@gmail.com"
__version__ = "0.10.0"

from .ARMAGARCH import ArmaGarch
from .Basic import BasicForecaster
from .DynamicRegressor import DynamicRegressor
from .EAT import EAT
from .Ridge2 import Ridge2Regressor
from .VAR import VAR


__all__ = [
    "ArmaGarch",
    "BasicForecaster",
    "DynamicRegressor",
    "EAT",
    "Ridge2Regressor",
    "VAR",
]



from .unimultivariate import (
    compute_output_dates,
    compute_result_df,
    get_frequency,
)
from .univariate import compute_y_ts, format_univariate_forecast
from .multivariate import compute_y_mts, format_multivariate_forecast
from .tscv_indices import get_tscv_indices

__all__ = [
    "compute_output_dates",
    "compute_y_ts",
    "format_univariate_forecast",
    "compute_y_mts",
    "format_multivariate_forecast",
    "compute_result_df",
    "get_frequency",
    "get_tscv_indices",
]



import pickle
import rpy2

try:
    import rpy2.robjects.packages as rpackages
    from rpy2.robjects.packages import importr
    from rpy2.robjects.vectors import FloatVector, StrVector
    from rpy2 import rinterface, robjects
    from rpy2.robjects import r, default_converter
    from rpy2.rinterface import RRuntimeWarning
    from rpy2.rinterface_lib import callbacks
    from rpy2.rinterface_lib.embedded import RRuntimeError
    import rpy2.robjects.conversion as cv
except ImportError as e:
    RPY2_ERROR_MESSAGE = str(e)
    RPY2_IS_INSTALLED = False
else:
    RPY2_IS_INSTALLED = True

USAGE_MESSAGE = """
This Python class is based on R package 'ahead' (https://techtonique.github.io/ahead/). 
You need to install R (https://www.r-project.org/) and rpy2 (https://pypi.org/project/rpy2/).

Then, install R package 'ahead' (if necessary): 
>> R -e 'install.packages("ahead", repos = https://techtonique.r-universe.dev)'    
"""

r["options"](warn=-1)


def _none2null(none_obj):
    return r("NULL")


none_converter = cv.Converter("None converter")
none_converter.py2rpy.register(type(None), _none2null)

required_packages = ["ahead"]  # list of required R packages

packages_to_install = [
    x for x in required_packages if not rpackages.isinstalled(x)
]

utils = importr("utils")
graphics = importr("graphics")
base = importr("base")

# print(f" required_packages: {required_packages} \n")
# print(f" packages_to_install: {packages_to_install} \n")
# print(f" len(packages_to_install): {len(packages_to_install)} \n")

if len(packages_to_install) > 0:
    base.options(
        repos=base.c(
            techtonique="https://techtonique.r-universe.dev",
            CRAN="https://cloud.r-project.org",
        )
    )
    utils.install_packages(
        StrVector(packages_to_install)
    )  # dependencies of dependencies nightmare...

# check R version
# print(f"R version: {utils.packageVersion('ahead')}")

FLOATVECTOR = FloatVector
AHEAD_PACKAGE = importr("ahead")
CHECK_PACKAGES = True


def DEEP_COPY(x):
    return pickle.loads(pickle.dumps(x, -1))


NONE_CONVERTER = none_converter



import numpy as np


def get_tscv_indices(
    n=25, p=0.8, initial_window=5, horizon=3, fixed_window=False
):
    """Generate indices to split data into training and test set.

    Parameters
    ----------
    n : int, input series length.

    p : float, proportion of data in the training set

    initial_window : int, initial number of consecutive values in each
                     training set sample

    horizon : int, number of consecutive values in test set sample

    fixed_window : boolean, fixed window or increasing window

    """

    # Initialization of indices -----

    indices = np.arange(n)
    train_test_indices = []
    hold_out_indices = []

    # train index
    min_index_train = 0
    max_index_train = initial_window

    # test index
    min_index_test = max_index_train
    max_index_test = initial_window + horizon

    # Main loop -----

    if p == 1:

        if fixed_window == True:

            while max_index_test <= n:

                train_test_indices.append(
                    {
                        "train": indices[min_index_train:max_index_train],
                        "test": indices[min_index_test:max_index_test],
                    }
                )

                min_index_train += 1
                min_index_test += 1
                max_index_train += 1
                max_index_test += 1

        else:  # fixed_window == False

            while max_index_test <= n:

                train_test_indices.append(
                    {
                        "train": indices[min_index_train:max_index_train],
                        "test": indices[min_index_test:max_index_test],
                    }
                )

                max_index_train += 1
                min_index_test += 1
                max_index_test += 1

        return train_test_indices

    # else if p < 1

    if fixed_window == True:

        while max_index_test <= n:

            if max_index_test <= int(p * n):

                train_test_indices.append(
                    {
                        "train": indices[min_index_train:max_index_train],
                        "test": indices[min_index_test:max_index_test],
                    }
                )

            else:

                hold_out_indices.append(
                    {
                        "train": indices[min_index_train:max_index_train],
                        "test": indices[min_index_test:max_index_test],
                    }
                )

            min_index_train += 1
            min_index_test += 1
            max_index_train += 1
            max_index_test += 1

    else:  # fixed_window == False

        while max_index_test <= n:

            if max_index_test <= int(p * n):

                train_test_indices.append(
                    {
                        "train": indices[min_index_train:max_index_train],
                        "test": indices[min_index_test:max_index_test],
                    }
                )

            else:

                hold_out_indices.append(
                    {
                        "train": indices[min_index_train:max_index_train],
                        "test": indices[min_index_test:max_index_test],
                    }
                )

            max_index_train += 1
            min_index_test += 1
            max_index_test += 1

    return train_test_indices, hold_out_indices



import numpy as np

from ..Base import Base
from ..utils import univariate as uv
from ..utils import unimultivariate as umv
from .. import config


class EAT(Base):
    """Combinations of ETS (exponential smoothing), auto.arima and Theta models

    Parameters:

        h: an integer;
            forecasting horizon

        level: an integer;
            Confidence level for prediction intervals

        weights: a list;
            coefficients assigned to each method in the ensemble

        type_pi: a string;
            Type of prediction interval (currently "gaussian",
            ETS: "E", Arima: "A" or Theta: "T")

        date_formatting: a string;
            Currently:
            - "original": yyyy-mm-dd
            - "ms": milliseconds

    Attributes:

        fcast_: an object;
            raw result from fitting R's `ahead::eatf` through `rpy2`

        averages_: a list;
            mean forecast in a list

        ranges_: a list;
            lower and upper prediction intervals in a list

        output_dates_: a list;
            a list of output dates (associated to forecast)

        mean_: a numpy array
            contains series mean forecast as a numpy array

        lower_: a numpy array
            contains series lower bound forecast as a numpy array

        upper_: a numpy array
            contains series upper bound forecast as a numpy array

        result_df_: a data frame;
            contains 3 columns, mean forecast, lower + upper
            prediction intervals, and a date index

    Examples:

    ```python
    import pandas as pd
    from ahead import EAT

    # Data frame containing the time series
    dataset = {
    'date' : ['2020-01-01', '2020-02-01', '2020-03-01', '2020-04-01', '2020-05-01'],
    'value' : [34, 30, 35.6, 33.3, 38.1]}

    df = pd.DataFrame(dataset).set_index('date')
    print(df)

    # univariate time series forecasting
    e1 = EAT(h = 5) # default, equal weights for each model=[1/3, 1/3, 1/3]
    e1.forecast(df)
    print(e1.result_df_)
    ```

    """

    def __init__(
        self,
        h=5,
        level=95,
        weights=None,
        type_pi="E",
        date_formatting="original",
    ):

        super().__init__(h=h, level=level)

        if weights is None:
            weights = [1 / 3, 1 / 3, 1 / 3]

        assert len(weights) == 3, "must have 'len(weights) == 3'"

        self.weights = weights
        self.type_pi = type_pi
        self.date_formatting = date_formatting
        self.input_df = None
        self.type_input = "univariate" 

        self.fcast_ = None
        self.averages_ = None
        self.ranges_ = None
        self.output_dates_ = []
        self.mean_ = []
        self.lower_ = []
        self.upper_ = []
        self.result_df_ = None

    def forecast(self, df):
        """Forecasting method from `EAT` class

        Parameters:

            df: a data frame;
                a data frame containing the input time series (see example)

        """

        # get input dates, output dates, number of series, series names, etc. 
        self.init_forecasting_params(df)

        # obtain time series object -----
        self.format_input()

        self.get_forecast("eat")

        # result -----
        (
            self.averages_,
            self.ranges_,
            _,
        ) = uv.format_univariate_forecast(
            date_formatting=self.date_formatting,
            output_dates=self.output_dates_,
            horizon=self.h,
            fcast=self.fcast_,
        )

        self.mean_ = np.asarray(self.fcast_.rx2["mean"])
        self.lower_ = np.asarray(self.fcast_.rx2["lower"])
        self.upper_ = np.asarray(self.fcast_.rx2["upper"])

        self.result_df_ = umv.compute_result_df(self.averages_, self.ranges_)

        return self



import numpy as np
import pandas as pd
from difflib import SequenceMatcher


# compute input dates from data frame's index
def compute_input_dates(df):
    
    input_dates = df.index.values

    frequency = pd.infer_freq(pd.DatetimeIndex(input_dates))

    input_dates = pd.date_range(
        start=input_dates[0], periods=len(input_dates), freq=frequency
    ).values.tolist()

    df_input_dates = pd.DataFrame({"date": input_dates})

    input_dates = pd.to_datetime(df_input_dates["date"]).dt.date
    
    return input_dates


# compute output dates from data frame's index
def compute_output_dates(df, horizon):
    input_dates = df.index.values

    if input_dates[0] == 0:
        input_dates = pd.date_range(
            start=pd.Timestamp.today().strftime("%Y-%m-%d"), periods=horizon
        )

    # print(f"\n in nnetsauce.utils.timeseries 1: {input_dates} \n")

    frequency = pd.infer_freq(pd.DatetimeIndex(input_dates))

    # print(f"\n in nnetsauce.utils.timeseries 2: {frequency} \n")

    output_dates = np.delete(
        pd.date_range(
            start=input_dates[-1], periods=horizon + 1, freq=frequency
        ).values,
        0,
    ).tolist()

    # print(f"\n in nnetsauce.utils.timeseries 3: {output_dates} \n")

    df_output_dates = pd.DataFrame({"date": output_dates})

    output_dates = pd.to_datetime(df_output_dates["date"]).dt.date    

    return output_dates, frequency


def compute_result_df(averages, ranges):
    pred_mean = pd.Series(dict(averages)).to_frame("mean")
    pred_ci = pd.DataFrame(
        ranges, columns=["date", "lower", "upper"]
    ).set_index("date")
    return pd.concat([pred_mean, pred_ci], axis=1)


def get_closest_str(input_str, list_choices):
    scores = np.asarray(
        [
            SequenceMatcher(None, a=input_str, b=elt).ratio()
            for elt in list_choices
        ]
    )
    return list_choices[np.where(scores == np.max(scores))[0][0]]


def get_frequency(input_str):
    # https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases
    """https://otexts.com/fpp2/ts-objects.html#frequency-of-a-time-series
    Data 	frequency
    Annual 	        1
    Quarterly 	    4
    Monthly 	    12
    Weekly 	        52
    """
    frequency_choices = {
        "A": 1,
        "Y": 1,
        "BA": 1,
        "BY": 1,
        "AS": 1,
        "YS": 1,
        "BAS": 1,
        "BYS": 1,
        "Q": 4,
        "BQ": 4,
        "QS": 4,
        "BQS": 4,
        "M": 12,
        "BM": 12,
        "CBM": 12,
        "MS": 12,
        "BMS": 12,
        "CBMS": 12,
        "W": 52,
        "B": 365,
        "C": 365,
        "D": 365,
    }

    try:

        return frequency_choices[input_str]

    except:

        list_frequency_choices = list(frequency_choices.keys())

        closest_str = get_closest_str(
            input_str=input_str, list_choices=list_frequency_choices
        )

        return frequency_choices[closest_str]



import os
import numpy as np
import pandas as pd

import rpy2.robjects as robjects
import rpy2.robjects.packages as rpackages
from rpy2.robjects.packages import importr
from rpy2.robjects import FloatVector
from datetime import datetime
from rpy2.robjects.vectors import StrVector
from .unimultivariate import get_frequency

base = importr("base")
stats = importr("stats")


def compute_y_ts(df, df_frequency):

    input_series = df.to_numpy()

    ts = stats.ts(
        FloatVector(input_series.flatten()),
        frequency=get_frequency(df_frequency),
    )

    return ts


def format_univariate_forecast(date_formatting, output_dates, horizon, fcast):

    if date_formatting == "original":
        output_dates_ = [
            datetime.strftime(output_dates[i], "%Y-%m-%d")
            for i in range(horizon)
        ]

    if date_formatting == "ms":
        output_dates_ = [
            int(
                datetime.strptime(str(output_dates[i]), "%Y-%m-%d").timestamp()
                * 1000
            )
            for i in range(horizon)
        ]

    averages = [
        [output_dates_[i], fcast.rx2["mean"][i]] for i in range(horizon)
    ]
    ranges = [
        [output_dates_[i], fcast.rx2["lower"][i], fcast.rx2["upper"][i]]
        for i in range(horizon)
    ]

    return averages, ranges, output_dates_



import os
import numpy as np
import pandas as pd

import rpy2.robjects as robjects
import rpy2.robjects.packages as rpackages
from rpy2.robjects.packages import importr
from rpy2.robjects import FloatVector, r
from datetime import datetime
from rpy2.robjects.vectors import StrVector
from .unimultivariate import get_frequency

base = importr("base")
stats = importr("stats")


def compute_y_mts(df, df_frequency):

    input_series = df.to_numpy()

    input_series_tolist = input_series.tolist()
    xx = [item for sublist in input_series_tolist for item in sublist]

    ts = r.matrix(
        FloatVector(xx),
        byrow=True,
        nrow=len(input_series_tolist),
        ncol=df.shape[1],
    )

    # ts.colnames = StrVector(df.columns.tolist())

    return stats.ts(ts, frequency=get_frequency(df_frequency))


def format_multivariate_forecast(
    n_series, date_formatting, output_dates, horizon, fcast
):
    if date_formatting == "original":
        output_dates_ = [
            datetime.strftime(output_dates[i], "%Y-%m-%d")
            for i in range(horizon)
        ]

    if date_formatting == "ms":
        output_dates_ = [
            int(
                datetime.strptime(str(output_dates[i]), "%Y-%m-%d").timestamp()
                * 1000
            )
            for i in range(horizon)
        ]

    mean_array = np.asarray(fcast.rx2["mean"], dtype="float64")
    lower_array = np.asarray(fcast.rx2["lower"], dtype="float64")
    upper_array = np.asarray(fcast.rx2["upper"], dtype="float64")

    averages = []
    ranges = []

    for j in range(n_series):
        averages_series_j = []
        ranges_series_j = []
        for i in range(horizon):
            date_i = output_dates_[i]
            averages_series_j.append([date_i, mean_array[i, j]])
            ranges_series_j.append(
                [
                    date_i,
                    lower_array[i, j],
                    upper_array[i, j],
                ]
            )
        averages.append(averages_series_j)
        ranges.append(ranges_series_j)

    return averages, ranges, output_dates_



import numpy as np

from ..Base import Base
from ..utils import multivariate as mv
from ..utils import unimultivariate as umv
from .. import config


class VAR(Base):
    """Vector AutoRegressive model

    Parameters:

        h: an integer;
            forecasting horizon

        level: an integer;
            Confidence level for prediction intervals

        lags: an integer;
            the lag order

        type_VAR: a string;
            Type of deterministic regressors to include
            ("const", "trend", "both", "none")

        date_formatting: a string;
            Currently:
            - "original": yyyy-mm-dd
            - "ms": milliseconds

    Attributes:

        fcast_: an object;
            raw result from fitting R's `ahead::varf` through `rpy2`

        averages_: a list of lists;
            mean forecast in a list for each series

        ranges_: a list of lists;
            lower and upper prediction intervals in a list for each series

        output_dates_: a list;
            a list of output dates (associated to forecast)

        mean_: a numpy array
            contains series mean forecast as a numpy array

        lower_: a numpy array
            contains series lower bound forecast as a numpy array

        upper_: a numpy array
            contains series upper bound forecast as a numpy array

        result_dfs_: a tuple of data frames;
            each element of the tuple contains 3 columns,
            mean forecast, lower + upper prediction intervals,
            and a date index

    Examples:

    ```python
    import pandas as pd
    from ahead import VAR

    # Data frame containing the time series
    dataset = {
    'date' : ['2001-01-01', '2002-01-01', '2003-01-01', '2004-01-01', '2005-01-01'],
    'series1' : [34, 30, 35.6, 33.3, 38.1],
    'series2' : [4, 5.5, 5.6, 6.3, 5.1],
    'series3' : [100, 100.5, 100.6, 100.2, 100.1]}
    df = pd.DataFrame(dataset).set_index('date')
    print(df)

    # multivariate time series forecasting
    v1 = VAR(h = 5, date_formatting = "original", type_VAR="none")
    v1.forecast(df)
    print(v1.result_dfs_)
    ```

    """

    def __init__(
        self, h=5, level=95, lags=1, type_VAR="none", date_formatting="original"
    ):  # type_VAR = "const", "trend", "both", "none"

        assert type_VAR in (
            "const",
            "trend",
            "both",
            "none",
        ), "must have: type_VAR in ('const', 'trend', 'both', 'none')"

        super().__init__(
            h=h,
            level=level,
        )

        self.lags = lags
        self.type_VAR = type_VAR
        self.date_formatting = date_formatting
        self.input_df = None

        self.fcast_ = None
        self.averages_ = None
        self.ranges_ = None
        self.output_dates_ = []
        self.mean_ = None
        self.lower_ = None
        self.upper_ = None
        self.result_dfs_ = None

    def forecast(self, df):
        """Forecasting method from `VAR` class

        Parameters:

            df: a data frame;
                a data frame containing the input time series (see example)

        """

        # get input dates, output dates, number of series, series names, etc. 
        self.init_forecasting_params(df)

        # obtain time series object -----
        self.format_input()

        self.get_forecast("var")

        # result -----
        (
            self.averages_,
            self.ranges_,
            _,
        ) = mv.format_multivariate_forecast(
            n_series=self.n_series,
            date_formatting=self.date_formatting,
            output_dates=self.output_dates_,
            horizon=self.h,
            fcast=self.fcast_,
        )

        self.mean_ = np.asarray(self.fcast_.rx2["mean"])
        self.lower_ = np.asarray(self.fcast_.rx2["lower"])
        self.upper_ = np.asarray(self.fcast_.rx2["upper"])

        self.result_dfs_ = tuple(
            umv.compute_result_df(self.averages_[i], self.ranges_[i])
            for i in range(self.n_series)
        )

        return self



import numpy as np
from .. import config

from ..utils import univariate as uv
from ..utils import unimultivariate as umv


class ArmaGarch(object):
    """ARMA(1, 1)-GARCH(1, 1) forecasting (with simulation)

    Parameters:

        h: an integer;
            forecasting horizon

        level: an integer;
            Confidence level for prediction intervals

        B: an integer;
            number of simulations for R's `stats::arima.sim`

        cl: an integer;
            the number of clusters for parallel execution (done in R /!\)

        dist: a string;
            distribution of innovations ("student" or "gaussian")

        seed: an integer;
            reproducibility seed

        date_formatting: a string;
            Currently:
            - "original": yyyy-mm-dd
            - "ms": milliseconds

    Attributes:

        fcast_: an object;
            raw result from fitting R's `ahead::armagarchf` through `rpy2`

        averages_: a list;
            mean forecast in a list

        ranges_: a list;
            lower and upper prediction intervals in a list

        output_dates_: a list;
            a list of output dates (associated to forecast)

        mean_: a numpy array
            contains series mean forecast as a numpy array

        lower_: a numpy array
            contains series lower bound forecast as a numpy array

        upper_: a numpy array
            contains series upper bound forecast as a numpy array

        result_df_: a data frame;
            contains 3 columns, mean forecast, lower + upper
            prediction intervals, and a date index

        sims_: a numpy array
            forecasting simulations

    """

    def __init__(
        self,
        h=5,
        level=95,
        B=250,
        cl=1,
        dist="student",
        seed=123,
        date_formatting="original",
    ):

        self.h = h
        self.level = level
        self.B = B
        self.cl = cl
        self.dist = dist
        self.seed = seed
        self.date_formatting = date_formatting
        self.input_df = None

        self.fcast_ = None
        self.averages_ = None
        self.ranges_ = None
        self.output_dates_ = []
        self.mean_ = []
        self.lower_ = []
        self.upper_ = []
        self.result_df_ = None
        self.sims_ = None

    def forecast(self, df):
        """Forecasting method from `ArmaGarch` class

        Parameters:

            df: a data frame;
                a data frame containing the input time series (see example)

        """

        # get input dates, output dates, number of series, series names, etc. 
        self.init_forecasting_params(df)        

        # obtain time series object -----
        self.format_input()

        self.get_forecast("armagarch")

        # result -----
        (
            self.averages_,
            self.ranges_,
            self.output_dates_,
        ) = uv.format_univariate_forecast(
            date_formatting=self.date_formatting,
            output_dates=self.output_dates_,
            horizon=self.h,
            fcast=self.fcast_,
        )

        self.mean_ = np.asarray(self.fcast_.rx2["mean"])
        self.lower_ = np.asarray(self.fcast_.rx2["lower"])
        self.upper_ = np.asarray(self.fcast_.rx2["upper"])

        self.result_df_ = umv.compute_result_df(self.averages_, self.ranges_)

        self.sims_ = np.asarray(self.fcast_.rx2["sims"])

        return self
